/*
   Programátor čipu ATF22V10C

   - příkazy se zadávají v textové podobě
     C [ROW] [FUSE] <CR>

   - seznam příkazů:
     I <CR>             pošle identifikaci hardwaru/firmwaru. Odpověď:
                        V název a verze programátoru
                        D název podporovaného prvku (každý zvlášť)
                        OK        

     D device           prvek, na který se má programátor přepnout. Odpoveď:
                        OK - programátor je připraven
                        E text chyby

     R row <CR>         načte jeden řádek pojistek z čipu. Odpověď:
                        R row 111011....110
                        OK

                        nebo

                        E text chyby

     P <CR>             přečte z čipu PES
                        PES TEXT
                        OK

     p                  vypne power down mód                   

     A <CR>             načte celý čip, tj. pojistky, pes, ues, konfiguraci. Odpověď:
                        P pes identifikátor
                        R 0 11100111...110
                        R 1 11100010...110
                        ...
                        U user identifikátor
                        C 101...1101            (konfigurace)
                        OK

                        nebo

                        E text chyby

     E <CR>             smaže celý čip. Odpověď:
                        OK
                        E text chyby

     W row fuses <CR>   zapíše na zadaný řádek hodnoty pojistek, provede i kontrolu zápisu. Odpověď:
                        OK
                        E text chyby

     U "text" <CR>      zapíše do uživatelské paměti 8 znakový text. Odpověď
                        OK
                        E text chyby
*/

#define FIRMWARE_VERSION "PLD programmer v1.0"    /* odesílá se při identifikaci */
#define CHIP "ATF22V10C"                          /* odesílá se při identifikaci + tento jediný čip je podporovaný příkazem D */

#define MAX_COMMAND_LENGTH 255

/* pin, kterým se zapíná 5V = napájení ATF + pin EDIT, log. 0 => zapne 5V */
#define V5_ENABLE_PIN 2
/* pin, kterým se zapíná 12V = aktivace programování na pinu EDIT (atfpin2), log. 1 => zapne 12V */
#define V12_ENABLE_PIN 11
/* atfpin3 P/V - programm/verify */
#define PV_PIN 4
/* atfpin4/5/6/7/9 ROW0 = r0/r1/r2/r3/r5 */
#define ROW0_PIN 5
/* atfpin8 ROW1 = r4 */
#define ROW1_PIN 6
/* atfpin10 CLOCK */
#define CLOCK_PIN 7
/* atfpin11 DATA_OUT - sériový výstup do ATF */
#define DATA_OUT_PIN 8
/* atfpin13 STROBE  */
#define STROBE_PIN 9
/* atfpin14 DATA_OUT - sériový vstup z ATF */
#define DATA_IN_PIN 10


#define pv(a) digitalWrite(PV_PIN, a)
#define row0(a) digitalWrite(ROW0_PIN, a)
#define row1(a) digitalWrite(ROW1_PIN, a)
#define clk(a) digitalWrite(CLOCK_PIN, a)
#define data_out(a) digitalWrite(DATA_OUT_PIN, a)
#define strobe(a) digitalWrite(STROBE_PIN, a)


// načtený příkaz, jeden znak vyhrazený pro zakončující 0
char command[MAX_COMMAND_LENGTH + 1];

// zápisový kurzor do command
uint8_t cmd_ptr = 0;

// identifikátor příkazu
char cmd_id;

// načtené číslo řádku
uint8_t cmd_row;

// index právě načítané pojistky
uint8_t cmd_fuse;

// příznak, že se mohou přeskočit mezery v příkazu
bool skipSpace;


// buffer pro 132 bitů jednoho řádku
uint8_t fuses[132];

// buffer pro pojistky, které se mají zkontrolovat se zpsaným řádkem
uint8_t vfuses[132];

// buffer pro 20 bitů konfigurace (OLMS konfigurace)
uint8_t configuration[20];

// načtená signatura
char pes[10];

// načtená uživatelská oblast
char ues[9];

// právě nastavený prvek
char device[20];

// zapne 5V větev
void on_5V()
{
  digitalWrite(V5_ENABLE_PIN, LOW);
}

// vypne 5V větev
void off_5V()
{
  digitalWrite(V5_ENABLE_PIN, HIGH);
}

// zapne 12V větev
void on_12V()
{
  digitalWrite(V12_ENABLE_PIN, HIGH);
}

// vypne 12V větev
void off_12V()
{
  digitalWrite(V12_ENABLE_PIN, LOW);
}


// sekvence vypínání
void turn_off()
{
  off_12V();
  delay(5);
  strobe(0);
  clk(0);
  pv(0);
  row0(0);
  row1(0);
  data_out(0);
  delay(5);
  off_5V();
  delay(5);
  pinMode(PV_PIN, INPUT);
  pinMode(ROW0_PIN, INPUT);
  pinMode(ROW1_PIN, INPUT);
  pinMode(CLOCK_PIN, INPUT);
  pinMode(STROBE_PIN, INPUT);
  pinMode(DATA_OUT_PIN, INPUT);

}

// sekvence zapnutí a aktivace EDIT režimu
void turn_on()
{
  pinMode(PV_PIN, OUTPUT);
  pinMode(ROW0_PIN, OUTPUT);
  pinMode(ROW1_PIN, OUTPUT);
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(STROBE_PIN, OUTPUT);
  pinMode(DATA_OUT_PIN, OUTPUT);

  on_5V();
  delay(5);
  strobe(1);
  clk(0);
  pv(0);
  row0(0);
  row1(0);
  data_out(0);
  delay(5);
  on_12V();
  delay(5);
}

int sendBit(int v)
{
  data_out(v);
  clk(1);
  clk(0);
}

int receiveBit()
{
  int v = digitalRead(DATA_IN_PIN);
  clk(1);
  clk(0);
  return v;
}

// sériově odešle do ATF číslo čteného řádku
void sendRowIndex(int r)
{
  for (int i = 0; i < 5; i++, r <<= 1)
  {
    data_out(r & 32);
    clk(1);
    clk(0);
  }
  data_out(r & 32);
  strobe(0);
  strobe(1);
  data_out(0);
}

// sériově odešle do ATF číslo čteného řádku + vyvolá zápis
void sendRowIndexAndWrite(int r)
{
  for (int i = 0; i < 5; i++, r <<= 1)
  {
    data_out(r & 32);
    clk(1);
    clk(0);
  }
  data_out(r & 32);
  delay(1);
  pv(1);
  delay(2);
  strobe(1);
  delay(10);
  strobe(0);
  delay(2);
  pv(0);
  data_out(0);
}


// Přečte "count" pojistek z řádku "row"
void readRow(int row, int count)
{
  if (count < 0)
    return;
  if (count > 132)
    count = 131;
  if (row < 0 || row > 63)
    return;

  delay(1);
  row0(0);
  row1(0);
  // zadání čísla řádku
  sendRowIndex(row);

  // načte bity
  for (int b = 0; b < count; b++)
  {
    fuses[b] = receiveBit();
  }

}

// načte identifikační text čipu
void readPES()
{
  row0(0);
  row1(0);
  // identifikátor je v řádku 58
  readRow(58, 132);
  // vytvoření řetězce
  for (int i = 0; i < 9; i++)
  {
    int v = 0;
    for (int j = 0; j < 8; j++)
    {
      v = (v >> 1) | (fuses[i * 8 + j] << 7);
    }
    pes[7 - i] = v;
  }
  pes[9] = 0;
}

// načte 20 bitů konfigurace (nastavení OLMC)
void readConfiguration()
{
  row0(0);
  row1(1);
  //
  strobe(0);
  strobe(1);
  // načte bity
  for (int b = 0; b < 20; b++)
  {
    configuration[b] = receiveBit();
  }

}

// načte uživatelský identifikátor
void readUES()
{
  row0(0);
  row1(0);
  // identifikátor je v řádku 44 v posledních 64 bitech
  readRow(44, 132);
  // vytvoření řetězce
  for (int i = 0; i < 8; i++)
  {
    int v = 0;
    for (int j = 0; j < 8; j++)
    {
      v = (v >> 1) | (fuses[i * 8 + j + 68] << 7);
      //v = (v << 1) | (fuses[i * 8 + j + 68]);
    }
    ues[i] = v;
  }
  ues[8] = 0;
}

// zapíše uživatelský identifikátor
void writeUES()
{
  row0(0);
  row1(0);
  // nejrpve se přeskočí 68 bitů, neboť UES je v posledních 64 bitech v řádku
  for (int i = 0; i < 68; i++)
  {
    data_out(1);
    clk(1);
    clk(0);
  }
  // odešle 64 bitů tvořící UES
  for (int i = 0; i < 8; i++)
  {
    int v = ues[i];
    for (int j = 0; j < 8; j++, v >>= 1)
    {
      data_out(v & 1);
      clk(1);
      clk(0);
    }
  }
  // ues je v řádku 44
  sendRowIndexAndWrite(44);
}

// zapíše pojistky z buferu fuses do řádku, jehož číslo je v parametru row
void writeRow(int row)
{
  row0(0);
  row1(0);
  // odeslání 132 pojistek
  for (int i = 0; i < 132; i++)
  {
    data_out(fuses[i]);
    clk(1);
    clk(0);
  }
  // odešle číslo řádku a provede zápis
  sendRowIndexAndWrite(row);
}

// zapíše jediný bit - používá se k vypnutí powerdown módu, příp. security
void writeBit(int row, int value)
{
  row0(0);
  row1(0);
  data_out(value);
  clk(1);
  clk(0);
  // odešle číslo řádku a provede zápis
  sendRowIndexAndWrite(row);
}

// zapíše pojistky z buferu fuses do 20 bitové konfigurace
void writeCfg()
{
  row0(0);
  row1(1);
  // odeslání 20 pojistek, ty jsou uložené v buffer fuses
  for (int i = 0; i < 19; i++)
  {
    data_out(fuses[i]);
    clk(1);
    clk(0);
  }
  data_out(fuses[19]);

  // zápisový impuls
  delay(1);
  pv(1);
  delay(2);
  strobe(1);
  delay(10);
  strobe(0);
  delay(2);
  pv(0);
  data_out(0);
}

bool skip_spaces()
{
  while (command[cmd_ptr] == ' ')
  {
    cmd_ptr++;
  }
  return command[cmd_ptr] == 0;
}


// načte z aktuální pozice příkazu číslo řádku. Pokud je neplatné, vrátí -1
int get_row()
{
  if (skip_spaces())
    return -1;
  int v = -1;
  while ( command[cmd_ptr] >= '0' && command[cmd_ptr] <= '9')
  {
    if ( v < 0)
      v = 0;
    v = v * 10 + command[cmd_ptr++] - '0';
  }
  return v > 63 ? -1 : v;
}


// načte pojistky - musí být min. 1, max. 132, jsou ve tvaru nepřerušovaného textu složeného ze znaků 1 a 0
bool get_fuses()
{
  if (skip_spaces())
    return false;
  for (int i = 0; i < 132; i++)
  {
    char z = command[cmd_ptr++];
    if ( z == 13 || z == 0)
    {
      cmd_ptr--;
      z = '1';
    }
    if ( z == '0' || z == '1')
    {
      fuses[i] = z - '0';
      vfuses[i] = z - '0';
    }
    else
    {
      // chyba
      return false;
    }
  }
  return true;
}

void send_row(int row)
{
  readRow(row, 132);
  Serial.print("R ");
  Serial.print(row);
  Serial.print(" ");
  // načte bity
  for (int b = 0; b < 132; b++)
  {
    Serial.write(48 + fuses[b]);
  }
  Serial.println();
}


void send_pes()
{
  readPES();
  Serial.print("P ");
  Serial.print(pes);
  Serial.println();
}


void send_ues()
{
  readUES();
  Serial.print("U ");
  Serial.print(ues);
  Serial.println();
}


void send_cfg()
{
  readConfiguration();
  Serial.print("C ");
  // načte bity
  for (int b = 0; b < 20; b++)
  {
    Serial.write(48 + configuration[b]);
  }
  Serial.println();
}

void erase()
{
  row0(1);
  row1(1);
  pv(1);
  strobe(0);
  delay(10);
  strobe(1);
  pv(0);
}


// provede příkaz v bufferu command
void processCommand()
{
  cmd_ptr = 0;
  skip_spaces();
  // příkaz - jeden znak
  char cmd = command[cmd_ptr++];
  switch (cmd)
  {
    case 'p':
      // vypne powerdown mód - zapíše na řádek PD 0
      turn_on();
      writeBit(59,0);
      turn_off();
      Serial.println("OK");
      break;
    case 'P':
      // odešle PES
      // inicializace
      turn_on();
      send_pes();
      turn_off();
      Serial.println("OK");
      break;

    case 'D':
      // přepne programátor na zadaný čip - teď jen kontrola, že se jedná o podporovaný ATF22V10C
      if (!skip_spaces())
      {
        for(int i = 0; i < sizeof(device) - 1; i++)
        {
          char c = command[cmd_ptr++];
          if(c == '\n' || c == '\r')
          {
            cmd_ptr--;
            c = 0;
          }
          device[i] = c;
        }
        device[sizeof(device)-1] = 0;
        if(strcmp(device, CHIP) == 0)
          Serial.println("OK");
        else
          Serial.println("E unsupported device!");
      }
      else 
      {
        device[0] = 0;  
        Serial.println("E device name missing!");
      }
      break;

    case 'I':
      // identifikace zařízení
      Serial.print("V ");
      Serial.println(FIRMWARE_VERSION);
      // výčet podporovaných čipů (zatím jen jeden)
      Serial.print("D ");
      Serial.println(CHIP);
      Serial.println("OK");
      break;
      
    case 'R':
      {
        // načte jeden řádek
        int row = get_row();
        if (row < 0)
        {
          Serial.println("E invalid row number");
        }
        else
        {
          turn_on();
          send_row(row);
          turn_off();
          Serial.println("OK");
        }
        break;
      }
      
    case 'W':
      {
        // zapíše a verifikuje jeden řádek
        int row = get_row();
        if (row < 0)
        {
          Serial.println("E invalid row number");
        }
        else
        {
          if (get_fuses())
          {
            turn_on();
            writeRow(row);
            // kontrola zápisu
            turn_off();
            delay(10);
            turn_on();
            readRow(row, 132);
            for (int i = 0; i < 132; i++)
            {
              if (fuses[i] != vfuses[i])
              {
                turn_off();
                // chyba zápisu
                Serial.println("E write error");
                return;
              }
            }
            turn_off();
            Serial.println("OK");

          }
          else
          {
            // chyba v načtených pojistkách
            Serial.println("E invalid fuses data");
          }
        }
        break;
      }
      
    case 'C':
      {
        // zapíše se konfigurace
        if (get_fuses())
        {
          turn_on();
          writeCfg();
          // kontrola zápisu
          turn_off();
          delay(2);
          turn_on();
          readConfiguration();
          for (int i = 0; i < 20; i++)
          {
            if (fuses[i] != configuration[i])
            {
              turn_off();
              // chyba zápisu
              Serial.println("E write error");
              return;
            }
          }
          turn_off();
          Serial.println("OK");
        }
        else
        {
          // chyba v načtených pojistkách
          Serial.println("E invalid fuses data");
        }

        break;
      }
    case 'A':
      {
        // pošle veškerý obsah čipu
        turn_on();
        send_pes();
        turn_off();
        turn_on();
        for (int row = 0; row < 44; row++)
        {
          send_row(row);
        }
        send_ues();
        send_cfg();
        turn_off();
        Serial.println("OK");
        break;
      }

    case 'E':
      {
        // smaže čip
        turn_on();
        erase();
        turn_off();
        Serial.println("OK");
        break;
      }

    case 'U':
      {
        if (skip_spaces())
        {
          // U bez parametru - načteme UES a pošleme
          turn_on();
          send_ues();
          turn_off();
        }
        else
        {
          // načte 8B textu
          for (int i = 0; i < 8; i++)
          {
            if (command[cmd_ptr] != 13)
            {
              ues[i] = command[cmd_ptr++];
            }
            else
              ues[i] = ' ';
          }
          ues[9] = 0;
          // zapíše UES
          turn_on();
          writeUES();
          turn_off();
          Serial.println("UES ");
          Serial.println(ues);
          Serial.println("OK");
        }
        break;
      }
    default:
      Serial.println("E invalid command");
      break;
  }
}

void setup() {
  // sériový port
  Serial.begin(115200);

  // vypnutí napěťových zdrojů
  off_5V(); // nutno udělat před módováním na výstup, jinak se krátkodobě 5V zapne (5V se zapíná v log. 0)
  off_12V();
  pinMode(V5_ENABLE_PIN, OUTPUT);
  pinMode(V12_ENABLE_PIN, OUTPUT);

  // vypnutí všeho
  turn_off();

}

void loop() {

  // posbírá znaky do jednoho příkazu, eliminuje netextové bajty a příliš dlouhé příkazy
  while (Serial.available())
  {
    int b = Serial.read();
    if (b == 10)
    {
      continue;
    }
    if (b == 13)
    {
      // příkaz je kompletní
      command[cmd_ptr] = 0;
      processCommand();
      cmd_ptr = 0;
      continue;
    }
    if ( b < 32 || b > 127)
    {
      // nepovolený znak, tohle je nějaký komunikační omyl
      cmd_ptr = 0;
      // vyčtou se zbylé znaky
      while (Serial.read() != -1);
      Serial.print("E syntax error '");
      Serial.print(b);
      Serial.println("'");
      // ukončení zpracování příkazů
      break;
    }
    // připíše znak příkazu do bufferu
    if (cmd_ptr < 255)
      command[cmd_ptr++] = b;
    else
    {
      // příliš dlouhý příkaz
      cmd_ptr = 0;
      Serial.println("E command too long");
    }
  }
}
